;***** Servo tester

;.device AT90S2313   automatic included in 2313def.inc 
.include "tn2313def.inc" 

.equ	TC0ReloadValue	=217	;(count 39 cycles)

;Timer 0 used for program timing purposes
;Timer 1 is free-running @ 10 MHz to generate and measure pulsewidths

.equ	LCD_PORT	=PORTB
.equ	LCD_RS		=PB0	;Portbit
.equ	LCD_RW		=PB1	;Portbit
.equ	LCD_E		=PB2	;Portbit
.equ	LCD_PosReadout	=0xC0

.equ	KEYBOARD_PORT	=PIND
.equ	KEY_UP		=PIND0
.equ	KEY_DOWN	=PIND1
.equ	KEY_LEFT	=PIND3
.equ	KEY_RIGHT	=PIND2
.equ	KEYBOARD_MASK	=(1<<KEY_UP)+(1<<KEY_DOWN)+(1<<KEY_LEFT)+(1<<KEY_RIGHT) 
.equ	KEY_DEBOUNCE	=5 
.equ	KEY_FIRSTREPEAT	=100 
.equ	KEY_OTHERREPEAT	=25 

.equ	INIT_PULSE_WIDTH=15200		; 1.5200 ms  Do not use leading 0's or the operand will be in octal
.equ	MAX_PULSE_WIDTH	=22000		; 2.2000 ms
.equ	MIN_PULSE_WIDTH	= 8000		; 0.8000 ms
.equ	INIT_SWEEP_MIN	=10000		; 1.0000 ms
.equ	INIT_SWEEP_MAX	=20000		; 2.0000 ms
.equ	INIT_SWEEP_STEP	=   10		; 0.0010 ms

.equ	MENU_OUT	=0
.equ	MENU_IN		=1
.equ	MENU_INAVG	=2
.equ	MENU_TRANS	=3
.equ	MENU_SWEEPMIN	=4
.equ	MENU_SWEEPMAX	=5
.equ	MENU_SWEEPSTEP	=6
.equ	MENU_SWEEPON	=7
.equ	MENU_LASTITEM	=MENU_SWEEPON

.equ	AVG_SIZEEXP	=5		;Averaging buffer size. 2^n elements
.equ	AVG_ELEMENTS	=(1<<AVG_SIZEEXP)	; = 2^5 = 32 elements (words)
.equ	AVG_BUFFERSIZE	=2*AVG_ELEMENTS		; = 64 bytes

.equ	INWATCH_TIME	=10		; = 10*10ms


.def	SaveSREG	=r1

.def	OutPulseL	=r2
.def	OutPulseH	=r3
.equ	OutPulseMemLoc	=2

.def	InPulseTempL	=r4
.def	InPulseTempH	=r5

.def	InPulseL	=r6
.def	InPulseH	=r7
.equ	InPulseMemLoc	=6

.def	MsCountL	=r8
.def	MsCountH	=r9
.def	LastKeyRead	=r10
.def	ChangeByL	=r11
.def	ChangeByH	=r12
.def	AvgBufferPtr	=r13
.def	InputWatchdog	=r14
.def	Menu		=r19
.def	CursorPos	=r20
.def	Keyboard	=r21
.def	Flags		=r22
  .equ	Flags_KeyRepeat	=0
  .equ	Flags_UpdateDisplay=1
  .equ	Flags_SweepDir	=2
  .equ	Flags_NewInput	=3
.def	KeyCount	=r23
.def	MsCount10	=r24
.def	State		=r25
; r26-31 reserved to X, Y and Z registers


.dseg
InAvgPulse:		.byte 2
SweepMinPulse:		.byte 2
SweepMaxPulse:		.byte 2
SweepStepPulse:		.byte 2

AvgSum:			.byte 3
AvgBuffer:		.byte AVG_BUFFERSIZE


.eseg
MenuOutText:		.db "Out     "
MenuInText:		.db "In      "
MenuInAvgText:		.db "In avg  "
MenuTransText:		.db "Transp. "
MenuSwpMinText:		.db "Swp min "
MenuSwpMaxText:		.db "Swp max "
MenuSwpStepText:	.db "SwpStep "
MenuSwpOnText:		.db "SweepOn "
InitText1a:		.db "ServoTes"
InitText1b:		.db "ter 2.05"
InitText2a:		.db "060823 e"
InitText2b:		.db "jberg.dk"


.cseg 
.org	0 
	rjmp	RESET     ; Reset Handler 
	rjmp	EXT_INT0  ; IRQ0 Handler 
	rjmp	EXT_INT1  ; IRQ1 Handler 
	rjmp	TIM_CAPT1 ; Timer1 Input Capture Handler 
	rjmp	TIM_COMP1 ; Timer1 Output Compare Handler 
	rjmp	TIM_OVF1  ; Timer1 Overflow Handler 
	rjmp	TIM_OVF0  ; Timer0 Overflow Handler 
	rjmp	UART_RXC  ; UART RX Complete Handler 
	rjmp	UART_DRE  ; UDR Empty Handler 
	rjmp	UART_TXC  ; UART TX Complete Handler 
	rjmp	ANA_COMP  ; Analog Comparator Handler 


;**** Interrupt routines 

EXT_INT0:  ; IRQ0 Handler
	reti


EXT_INT1:  ; IRQ1 Handler
	reti


TIM_CAPT1: ; Timer1 Input Capture Handler 
	in	SaveSREG,SREG 
	push	r16 
	push	r17 
	in	r16,TCCR1B		;Reverse edge trigger on Input Capture 
	ldi	r17,(1<<ICES1)		;to capture both edges
	eor	r16,r17 
	out	TCCR1B,r16 
	sbrc	r16,ICES1 
	rjmp	TIM_CAPT1_PULSEEND 
	in	InPulseTempL,ICR1L	;Positive edge. Save captured value
	in	InPulseTempH,ICR1H 
	rjmp	TIM_CAPT1_EXIT 
TIM_CAPT1_PULSEEND:
	in	InPulseL,ICR1L		;Negative edge. Get captured value
	in	InPulseH,ICR1H
	sub	InPulseL,InPulseTempL	;Find difference from positive edge
	sbc	InPulseH,InPulseTempH	;to get pulsewidth
	push	ZL
	push	ZH
	ldi	ZL,LOW(InPulseMemLoc)
	ldi	ZH,HIGH(InPulseMemLoc)
	rcall	ForceValidPulseWidth	;Check for valid pulsewidth
	pop	ZH
	pop	ZL
	brts	TIM_CAPT1_EXIT		;Exit if input pulse is invalid
	sbr	Flags,(1<<Flags_UpdateDisplay)+(1<<Flags_NewInput)	;Set flags
	ldi	r16,INWATCH_TIME
	mov	InputWatchdog,r16	;Reset valid input monitor
	cpi	Menu,MENU_TRANS
	brne	TIM_CAPT1_EXIT
	mov	OutPulseL,InPulseL	;If menu "Transparent" selected
	mov	OutPulseH,InPulseH	;then copy input to output
TIM_CAPT1_EXIT:
	pop	r17
	pop	r16
	out	SREG,SaveSREG
	reti


TIM_COMP1: ; Timer1 Output Compare Handler
	reti


TIM_OVF1:  ; Timer1 Overflow Handler
	in	SaveSREG,SREG
	push	r16
	cpi	State,0
	breq	TIM_OVF1_STATE0		;Find current pulsewidth generator state
	cpi	State,1
	breq	TIM_OVF1_STATE1
	cpi	State,2
	breq	TIM_OVF1_STATE2
	rjmp	TIM_OVF1_EXIT
TIM_OVF1_STATE0:			;State 0: Preparing for generation of new pulse
	inc	State			;At next TC1 overflow, the positive edge will be generated by hardware
	clr	r16
	out	OCR1AH,r16		;Set compare value to 0
	out	OCR1AL,r16
	ldi	r16,(1<<COM1A1)+(1<<COM1A0)
	out	TCCR1A,r16		;Set OC to 1 at compare match
	rjmp	TIM_OVF1_EXIT
TIM_OVF1_STATE1:			;State 1: New pulse just started. Set up hardware for negative edge gen.
	inc	State
	out	OCR1AH,OutPulseH	;Set compare value to desired pulse width
	out	OCR1AL,OutPulseL
	ldi	r16,(1<<COM1A1)
	out	TCCR1A,r16		;Set OC to 0 at compare match
	rjmp	TIM_OVF1_EXIT
TIM_OVF1_STATE2:			;State 2: Delay before next pulse
	clr	State
TIM_OVF1_EXIT:
	pop	r16
	out	SREG,SaveSREG
	reti


TIM_OVF0:  ; Timer0 Overflow Handler
	in	SaveSREG,SREG		;Called every 1 ms (exactly 0.9984 ms at 10.000 MHz xtal)
	push	r16
	ldi	r16,TC0ReloadValue
	out	TCNT0,r16
	tst	MsCountL		;MsCount[H:L] used for delays in mainprog
	breq	TIM_OVF0_LOW0		;If contents are other than 0, decrease by 1 each ms
	dec	MsCountL
	rjmp	TIM_OVF0_CHECKMS10
TIM_OVF0_LOW0:
	tst	MsCountH
	breq	TIM_OVF0_CHECKMS10
	dec	MsCountL
	dec	MsCountH
TIM_OVF0_CHECKMS10:
	dec	MsCount10
	brne	TIM_OVF0_EXIT
	ldi	MsCount10,10
	rcall	Int_CheckKeyboard_r16	;Executed every 10 ms
	rcall	Int_CheckSweep_r16
	rcall	Int_CheckInput_r16
TIM_OVF0_EXIT: 
	pop	r16 
	out	SREG,SaveSREG 
	reti 

Int_CheckKeyboard_r16:			;Subfunction for Timer0 Overflow Handler
	in	r16,KEYBOARD_PORT	;Scans keyboard for activity
	com	r16			;Keyboard inputs active low
	andi	r16,KEYBOARD_MASK
	cp	r16,LastKeyRead
	breq	ICK_SameKey
	mov	LastKeyRead,r16		;Debouncing
	ldi	KeyCount,KEY_DEBOUNCE
	cbr	Flags,(1<<Flags_KeyRepeat)
	rjmp	ICK_Exit
ICK_SameKey:
	dec	KeyCount		;Keep counting for possible bounce
	brne	ICK_Exit
	mov	Keyboard,r16		;New keypress available
	sbrc	Flags,Flags_KeyRepeat
	rjmp	ICK_KeyRepeat
	ldi	KeyCount,KEY_FIRSTREPEAT	;Key just pressed. Set up long key repeat time
	sbr	Flags,(1<<Flags_KeyRepeat)
	rjmp	ICK_Exit
ICK_KeyRepeat:
	ldi	KeyCount,KEY_OTHERREPEAT	;Key still held down. Set up short key repeat time
ICK_Exit:
	ret

Int_CheckSweep_r16:			;Subfunction for Timer0 Overflow Handler
	cpi	Menu,MENU_SWEEPON	;Updates generated pulsewidth if in sweepmode
	brne	ICS_Exit
	push	YL
	push	YH
	sbr	Flags,(1<<Flags_UpdateDisplay)
	sbrc	Flags,Flags_SweepDir
	rjmp	ICS_SweepDown
	lds	YL,SweepStepPulse
	lds	YH,SweepStepPulse+1
	add	OutPulseL,YL		;Increase pulsewidth
	adc	OutPulseH,YH
	lds	YL,SweepMaxPulse
	lds	YH,SweepMaxPulse+1
	cp	YL,OutPulseL		;Check for upper boundary
	cpc	YH,OutPulseH
	brsh	ICS_PopExit
	mov	OutPulseL,YL		;Upper boundary reached
	mov	OutPulseH,YH
	sbr	Flags,(1<<Flags_SweepDir)	;Set flag for counting down
	rjmp	ICS_PopExit
ICS_SweepDown:
	lds	YL,SweepStepPulse
	lds	YH,SweepStepPulse+1
	sub	OutPulseL,YL		;Decrease pulsewidth
	sbc	OutPulseH,YH
	lds	YL,SweepMinPulse
	lds	YH,SweepMinPulse+1
	cp	OutPulseL,YL		;Check for lower boundary
	cpc	OutPulseH,YH
	brsh	ICS_PopExit
	mov	OutPulseL,YL		;Lower boundary reached
	mov	OutPulseH,YH
	cbr	Flags,(1<<Flags_SweepDir)	;Set flag for counting up
ICS_PopExit:
	pop	YH
	pop	YL
ICS_Exit:
	ret

Int_CheckInput_r16:		;Subfunction for Timer0 Overflow Handler
	dec	InputWatchdog	;Monitors activity on input pulse
	brne	ICI_Exit	;If no valid input pulse has been received for 100ms:
	clr	InPulseL	;Simulate input pulse of 0.0000 ms
	clr	InPulseH
	sbr	Flags,(1<<Flags_UpdateDisplay)+(1<<Flags_NewInput)
	ldi	r16,INWATCH_TIME
	mov	InputWatchdog,r16
ICI_Exit:
	ret



 
UART_RXC:  ; UART RX Complete Handler 
	reti 

	 
UART_DRE:  ; UDR Empty Handler 
	reti 

	 
UART_TXC:  ; UART TX Complete Handler 
	reti 

	 
ANA_COMP:  ; Analog Comparator Handler 
	reti 
	 


;**** Procedures 

Delay_us:				;Delays execution for r16 us (+0.8us) 
	push	r16			;Only valid for 10.000 MHz xtal 
DelayUsLoop: 
	nop 
	nop 
	nop 
	nop 
	nop 
	nop 
	nop 
	dec	r16 
	brne	DelayUsLoop 
	pop	r16 
	ret 


Delay_ms:				;Delays execution for XH:XL ms (+ between 0 and 1 ms) 
	adiw	XL,1 
	cli 
	mov	MsCountL,XL		;MsCount[H:L] decrements by 1 each ms at TC0 overflow int. 
	mov	MsCountH,XH 
	sei 
DelayMsLoop: 
	tst	MsCountH 
	brne	DelayMsLoop 
	tst	MsCountL 
	brne	DelayMsLoop 
	ret 


LCD_SendHighNibble_r16_r17:		;Nibble in r16 
	andi	r16,0b11110000 
	sbi	LCD_PORT,LCD_E		;Set E 
	in	r17,LCD_PORT 
	andi	r17,0b00001111 
	or	r17,r16 
	out	LCD_PORT,r17 
	cbi	LCD_PORT,LCD_E		;Clear E 
	ret 


LCD_LocalSend: 
	push	r17 
	push	r16 
	rcall	LCD_SendHighNibble_r16_r17 
	ldi	r16,1 
	rcall	Delay_us		;1 us delay 
	pop	r16 
	push	r16 
	swap	r16 
	rcall	LCD_SendHighNibble_r16_r17 
	ldi	r16,40 
	rcall	Delay_us		;Delay 40 us 
	pop	r16 
	pop	r17 
	ret 


LCD_SendCommand:			;Commandbyte in r16 
	cbi	LCD_PORT,LCD_RS		;Clear RS 
	cbi	LCD_PORT,LCD_RW		;Clear R/W (Write) 
	rjmp	LCD_LocalSend 


LCD_SendData:				;Databyte in r16 
	sbi	LCD_PORT,LCD_RS		;Set RS 
	cbi	LCD_PORT,LCD_RW		;Clear R/W (Write) 
	rjmp	LCD_LocalSend 


LCD_Init: 
	push	r16 
	push	r17 
	ldi	XL,15 
	clr	XH 
	rcall	Delay_ms		;Delay 15 ms 
	ldi	r16,0x30 
	rcall	LCD_SendHighNibble_r16_r17 
	ldi	XL,5
	clr	XH
	rcall	Delay_ms		;Delay 5 ms
	ldi	r16,0x30 
	rcall	LCD_SendHighNibble_r16_r17 
	ldi	r16,100 
	rcall	Delay_us		;Delay 100 us 
	ldi	r16,0x30 
	rcall	LCD_SendHighNibble_r16_r17 
	ldi	r16,40 
	rcall	Delay_us		;Delay 40 us 
	ldi	r16,0x20 
	rcall	LCD_SendHighNibble_r16_r17 
	ldi	r16,40 
	rcall	Delay_us		;Delay 40 us 
	ldi	r16,0x28		;Init1: 4bit / 2 line / 5x7 dot format 
	rcall	LCD_SendCommand 	; (Why 2 lines? A 1x16 LCD is treated as a 2x8)
	ldi	r16,0x06		;Init2: Entry mode set / Increment / No shift 
	rcall	LCD_SendCommand 
	ldi	r16,0x0C		;Init3: Display ctrl. / Disp ON / Cursor OFF / Blink OFF
	rcall	LCD_SendCommand 
	ldi	r16,0x01		;Init4: Display Clear 
	rcall	LCD_SendCommand 
	ldi	XL,2 
	clr	XH 
	rcall	Delay_ms		;Delay 2 ms 
	ldi	r16,0x80		;Init5: DD RAM Address Set / 1st char
	rcall	LCD_SendCommand 
	pop	r17 
	pop	r16 
	ret 


FindDigit_r16_X:
	clr	r16		;Returns X DIV Y in r16
FD_Loop:			;and X MOD Y in X
	cp	XL,YL
	cpc	XH,YH
	brlo	FD_Exit
	inc	r16
	sub	XL,YL
	sbc	XH,YH
	rjmp	FD_Loop
FD_Exit:
	ret


DisplayPulseWidth:		;Pulsewidth to display in X
	push	r16
	ldi	YL,LOW(10000) 
	ldi	YH,HIGH(10000) 
	rcall	FindDigit_r16_X 
	ori	r16,'0' 
	rcall	LCD_SendData 
	ldi	r16,'.' 
	rcall	LCD_SendData 
	ldi	YL,LOW(1000) 
	ldi	YH,HIGH(1000) 
	rcall	FindDigit_r16_X 
	ori	r16,'0' 
	rcall	LCD_SendData 
	ldi	YL,LOW(100) 
	ldi	YH,HIGH(100) 
	rcall	FindDigit_r16_X 
	ori	r16,'0' 
	rcall	LCD_SendData 
	ldi	YL,LOW(10) 
	ldi	YH,HIGH(10) 
	rcall	FindDigit_r16_X 
	ori	r16,'0' 
	rcall	LCD_SendData 
	mov	r16,XL 
	ori	r16,'0' 
	rcall	LCD_SendData 
	ldi	r16,'m' 
	rcall	LCD_SendData 
	ldi	r16,'s' 
	rcall	LCD_SendData 
	pop	r16 
	ret 


DisplayText8_r1617_Z:		;Text pointer in ZL
	ldi	r17,8
DT_Loop:
	out	EEAR,ZL
	sbi	EECR,EERE
	inc	ZL
	in	r16,EEDR
	rcall	LCD_SendData
	dec	r17
	brne	DT_Loop
	ret


FetchAndDisplayPulseWidth:
	ldi	r16,LCD_PosReadout
	rcall	LCD_SendCommand	;Display position
	cli
	ld	XL,Z+		;Get pulsewidth from Z pointer
	ld	XH,Z
	sei
	rjmp	DisplayPulseWidth


ForceValidPulseWidth:
	push	XL		;Pulsewidth location to force in Z
	push	XH		;Can be called from an interrupt
	push	YL		;Returns T=1 if outside valid range
	push	YH
	clt
	ld	XL,Z+
	ld	XH,Z+
	ldi	YL,LOW(MAX_PULSE_WIDTH)
	ldi	YH,HIGH(MAX_PULSE_WIDTH)
	cp	YL,XL
	cpc	YH,XH
	brsh	FVPW_NotHigh
	mov	XL,YL
	mov	XH,YH
	set
	rjmp	FVPW_Exit
FVPW_NotHigh:
	ldi	YL,LOW(MIN_PULSE_WIDTH)
	ldi	YH,HIGH(MIN_PULSE_WIDTH)
	cp	XL,YL
	cpc	XH,YH
	brsh	FVPW_Exit
	mov	XL,YL
	mov	XH,YH
	set
FVPW_Exit:
	st	-Z,XH		;Save trimmed pulsewidth
	st	-Z,XL		;and restore Z
	pop	YH
	pop	YL
	pop	XH
	pop	XL
	ret


DisplayOut:			;Displays menu text
	ldi	ZL,MenuOutText
	rcall	DisplayText8_r1617_Z
	;Display outgoing pulsewidth
	ldi	ZL,LOW(OutPulseMemLoc)
	ldi	ZH,HIGH(OutPulseMemLoc)
	rjmp	FetchAndDisplayPulseWidth


DisplayIn:			;Displays menu text
	ldi	ZL,MenuInText
	rcall	DisplayText8_r1617_Z
	;Display incoming pulsewidth
	ldi	ZL,LOW(InPulseMemLoc)
	ldi	ZH,HIGH(InPulseMemLoc)
	rjmp	FetchAndDisplayPulseWidth


DisplayInAvg:			;Displays menu text
	ldi	ZL,MenuInAvgText
	rcall	DisplayText8_r1617_Z
	;Display averaged incoming pulsewidth
	ldi	ZL,LOW(InAvgPulse)
	ldi	ZH,HIGH(InAvgPulse)
	rjmp	FetchAndDisplayPulseWidth


DisplayTrans:			;Displays menu text
	ldi	ZL,MenuTransText
	rcall	DisplayText8_r1617_Z
	;Display outgoing pulsewidth
	ldi	ZL,LOW(OutPulseMemLoc)
	ldi	ZH,HIGH(OutPulseMemLoc)
	rjmp	FetchAndDisplayPulseWidth


DisplaySweepMin:		;Displays menu text
	ldi	ZL,MenuSwpMinText
	rcall	DisplayText8_r1617_Z
	;Display averaged incoming pulsewidth
	ldi	ZL,LOW(SweepMinPulse)
	ldi	ZH,HIGH(SweepMinPulse)
	rjmp	FetchAndDisplayPulseWidth


DisplaySweepMax:		;Displays menu text
	ldi	ZL,MenuSwpMaxText
	rcall	DisplayText8_r1617_Z
	;Display averaged incoming pulsewidth
	ldi	ZL,LOW(SweepMaxPulse)
	ldi	ZH,HIGH(SweepMaxPulse)
	rjmp	FetchAndDisplayPulseWidth


DisplaySweepStep:		;Displays menu text
	ldi	ZL,MenuSwpStepText
	rcall	DisplayText8_r1617_Z
	;Display averaged incoming pulsewidth
	ldi	ZL,LOW(SweepStepPulse)
	ldi	ZH,HIGH(SweepStepPulse)
	rjmp	FetchAndDisplayPulseWidth


DisplaySweepOn:			;Displays menu text
	ldi	ZL,MenuSwpOnText
	rcall	DisplayText8_r1617_Z
	;Display outgoing pulsewidth
	ldi	ZL,LOW(OutPulseMemLoc)
	ldi	ZH,HIGH(OutPulseMemLoc)
	rjmp	FetchAndDisplayPulseWidth



UpdateDisplay:				;Refreshes display
	push	r0
	push	r16
	ldi	r16,0x0C		;Display ctrl. / Disp ON / Cursor OFF / Blink OFF
	rcall	LCD_SendCommand 
	ldi	r16,0x80		;Start of first character
	rcall	LCD_SendCommand		;Display position
	ldi	ZL,LOW(UpdateDisplayJumpTable)
	ldi	ZH,HIGH(UpdateDisplayJumpTable)
	clr	r16
	add	ZL,Menu
	adc	ZH,r16
	icall				;Display appropriate menu
	;Set cursor at correct position
	ldi	ZL,LOW(UpdateDisplayCursorTable*2)
	ldi	ZH,HIGH(UpdateDisplayCursorTable*2)
	clr	r16
	add	ZL,CursorPos		;Find cursor position
	adc	ZH,r16
	lpm
	mov	r16,r0
	rcall	LCD_SendCommand
	tst	CursorPos
	brne	UD_BlinkOn
	ldi	ZL,LOW(UpdateDisplayCursorType*2)
	ldi	ZH,HIGH(UpdateDisplayCursorType*2)
	clr	r16
	add	ZL,Menu			;Find cursor type (menu and position dependant)
	adc	ZH,r16
	lpm
	mov	r16,r0
	rjmp	UD_SendCommand
UD_BlinkOn:
	ldi	r16,0x0D		;Display ctrl. / Disp ON / Cursor OFF / Blink ON
UD_SendCommand:	
	rcall	LCD_SendCommand		;Cursor type
	pop	r16
	pop	r0
	ret
UpdateDisplayJumpTable:
	rjmp	DisplayOut
	rjmp	DisplayIn
	rjmp	DisplayInAvg
	rjmp	DisplayTrans
	rjmp	DisplaySweepMin
	rjmp	DisplaySweepMax
	rjmp	DisplaySweepStep
	rjmp	DisplaySweepOn
UpdateDisplayCursorTable:
	.db 	0x87,LCD_PosReadout,LCD_PosReadout+2,LCD_PosReadout+3,LCD_PosReadout+4,LCD_PosReadout+5
UpdateDisplayCursorType:
	.db	0x0E,0x0C,0x0C,0x0C,0x0E,0x0E,0x0E,0x0C


UpdateChangeBy:
	push	r0
	ldi	ZL,LOW(UpdateChangeByTable*2)
	ldi	ZH,HIGH(UpdateChangeByTable*2)
	mov	r0,CursorPos
	lsl	r0		; *2
	add	ZL,r0
	clr	r0
	adc	ZH,r0
	lpm
	mov	ChangeByL,r0
	adiw	ZL,1
	lpm
	mov	ChangeByH,r0
	pop	r0
	ret
UpdateChangeByTable:
	.dw	0,10000,1000,100,10,1


CursorLeft:
	tst	CursorPos		;User tries to move the cursor left
	breq	CL_Exit
	dec	CursorPos
	rcall	UpdateChangeBy
CL_Exit:
	ret


CursorRight:
	cpi	CursorPos,5		;User tries to move the cursor right
	brsh	CR_Exit
	inc	CursorPos
	rcall	UpdateChangeBy
CR_Exit:
	ret


NA:
	ret				;Keypress with no function


PrevMenu:
	tst	Menu			;Move one menuentry back
	breq	PM_Exit
	dec	Menu
	clr	CursorPos
	rcall	UpdateChangeBy
PM_Exit:
	ret


NextMenu:
	cpi	Menu,MENU_LASTITEM	;Move one menuentry forward
	brsh	NM_Exit
	inc	Menu
	clr	CursorPos
	rcall	UpdateChangeBy
NM_Exit:
	ret


KeyUpOut:			;Key UP pressed in the OUT menu
	tst	CursorPos
	breq	KUO_Exit
	rjmp	IncreaseOutput	;Cursor is in pulsewidth editing. Increase pulsewidth
KUO_Exit:
	rjmp	PrevMenu	;Cursor is in menu selection. Find previous menu


KeyDownOut:			;Key DOWN pressed in the OUT menu
	tst	CursorPos
	breq	KDO_Menu
	rjmp	DecreaseOutput	;Cursor is in pulsewidth editing. Decrease pulsewidth
KDO_Menu:
	rjmp	NextMenu	;Cursor is in menu selection. Find next menu


IncreaseMemLoc:			;Takes a word in RAM @Z, and increases it by ChangeBy[H:L]
	ld	XL,Z+		;Returns new value in X
	ld	XH,Z+		;Must keep Z untouched
	add	XL,ChangeByL
	adc	XH,ChangeByH
	brcc	IML_NoWrap
	ser	XL		;Passed max
	ser	XH
IML_NoWrap:
	cli
	st	-Z,XH
	st	-Z,XL
	sei
	ret
	
DecreaseMemLoc:			;Takes a word in RAM @Z, and decreases it by ChangeBy[H:L]
	ld	XL,Z+		;Returns new value in X
	ld	XH,Z+		;Must keep Z untouched
	sub	XL,ChangeByL
	sbc	XH,ChangeByH
	brcc	DML_NoWrap
	clr	XL		;Passed 0
	clr	XH
DML_NoWrap:
	cli
	st	-Z,XH
	st	-Z,XL
	sei
	ret
	
IncreaseOutput:				;Increases output pulsewidth
	ldi	ZL,LOW(OutPulseMemLoc)
	ldi	ZH,HIGH(OutPulseMemLoc)
	rcall	IncreaseMemLoc
	cli				;OutPulse used by interrupt
	rcall	ForceValidPulseWidth	;Check for valid pulsewidth
	sei
	ret

DecreaseOutput:				;Decreases output pulsewidth
	ldi	ZL,LOW(OutPulseMemLoc)
	ldi	ZH,HIGH(OutPulseMemLoc)
	rcall	DecreaseMemLoc
	cli				;OutPulse used by interrupt
	rcall	ForceValidPulseWidth	;Check for valid pulsewidth
	sei
	ret

IncreaseSweepMin:
	ldi	ZL,LOW(SweepMinPulse)
	ldi	ZH,HIGH(SweepMinPulse)
	rcall	IncreaseMemLoc
	rcall	ForceValidPulseWidth	;Check for valid pulsewidth
	ldi	ZL,LOW(SweepMaxPulse)	;Compare to SweepMax and increase that as well if exceeded
	ldi	ZH,HIGH(SweepMaxPulse)	; (SweepMin may not be higher than SweepMax)
	ld	YL,Z+
	ld	YH,Z+
	cp	YL,XL
	cpc	YH,XH
	brsh	ISM_Exit
	st	-Z,XH			;Increase max to new min
	st	-Z,XL
ISM_Exit:
	ret

DecreaseSweepMin:
	ldi	ZL,LOW(SweepMinPulse)
	ldi	ZH,HIGH(SweepMinPulse)
	rcall	DecreaseMemLoc
	rcall	ForceValidPulseWidth	;Check for valid pulsewidth
	ret

IncreaseSweepMax:
	ldi	ZL,LOW(SweepMaxPulse)
	ldi	ZH,HIGH(SweepMaxPulse)
	rcall	IncreaseMemLoc
	rcall	ForceValidPulseWidth	;Check for valid pulsewidth
	ret

DecreaseSweepMax:
	ldi	ZL,LOW(SweepMaxPulse)
	ldi	ZH,HIGH(SweepMaxPulse)
	rcall	DecreaseMemLoc
	rcall	ForceValidPulseWidth	;Check for valid pulsewidth
	ldi	ZL,LOW(SweepMinPulse)	;Compare to min and decrease min if exceeded
	ldi	ZH,HIGH(SweepMinPulse)
	ld	YL,Z+
	ld	YH,Z+
	cp	XL,YL
	cpc	XH,YH
	brsh	DSM_Exit
	st	-Z,XH
	st	-Z,XL
DSM_Exit:
	ret

IncreaseSweepStep:
	ldi	ZL,LOW(SweepStepPulse)
	ldi	ZH,HIGH(SweepStepPulse)
	rcall	IncreaseMemLoc
	ret				;Do NOT check pulsewidth here (possible small values)

DecreaseSweepStep:
	ldi	ZL,LOW(SweepStepPulse)
	ldi	ZH,HIGH(SweepStepPulse)
	rcall	DecreaseMemLoc
	ret				;Do NOT check pulsewidth here (possible small values)


KeyUpSwMin:
	tst	CursorPos
	breq	KUSMi_Menu
	rjmp	IncreaseSweepMin
KUSMi_Menu:
	rjmp	PrevMenu


KeyDownSwMin:
	tst	CursorPos
	breq	KDSMi_Menu
	rjmp	DecreaseSweepMin
KDSMi_Menu:
	rjmp	NextMenu


KeyUpSwMax:
	tst	CursorPos
	breq	KUSMa_Menu
	rjmp	IncreaseSweepMax
KUSMa_Menu:
	rjmp	PrevMenu


KeyDownSwMax:
	tst	CursorPos
	breq	KDSMa_Menu
	rjmp	DecreaseSweepMax
KDSMa_Menu:
	rjmp	NextMenu


KeyUpSwStep:
	tst	CursorPos
	breq	KUSS_Menu
	rjmp	IncreaseSweepStep
KUSS_Menu:
	rjmp	PrevMenu


KeyDownSwStep:
	tst	CursorPos
	breq	KDSS_Menu
	rjmp	DecreaseSweepStep
KDSS_Menu:
	rjmp	NextMenu



;Each entry in the following tables corresponds to each menu,
;and the entry defines which function to call, when a key is pressed

KeyUpTable:
	rjmp	KeyUpOut
	rjmp	PrevMenu
	rjmp	PrevMenu
	rjmp	PrevMenu
	rjmp	KeyUpSwMin
	rjmp	KeyUpSwMax
	rjmp	KeyUpSwStep
	rjmp	PrevMenu

KeyDownTable:
	rjmp	KeyDownOut
	rjmp	NextMenu
	rjmp	NextMenu
	rjmp	NextMenu
	rjmp	KeyDownSwMin
	rjmp	KeyDownSwMax
	rjmp	KeyDownSwStep
	rjmp	NA
	
KeyLeftTable:
	rjmp	CursorLeft
	rjmp	NA
	rjmp	NA
	rjmp	NA
	rjmp	CursorLeft
	rjmp	CursorLeft
	rjmp	CursorLeft
	rjmp	NA

KeyRightTable:
	rjmp	CursorRight
	rjmp	NA
	rjmp	NA
	rjmp	NA
	rjmp	CursorRight
	rjmp	CursorRight
	rjmp	CursorRight
	rjmp	NA

ExecuteKeyCommand_Z:
	push	r16		;Finds correct function to call
	clr	r16		;Each menu has/can have it's own usage of the keyboard
	add	ZL,Menu
	adc	ZH,r16
	icall
	pop	r16
	ret


UpdateAverageBuffer:
	cli
	mov	XL,InPulseL		;First save InPulse
	mov	XH,InPulseH
	sei
	push	r0
	push	r16
	push	r17
	push	r18
	ldi	ZL,LOW(AvgSum)
	ldi	ZH,HIGH(AvgSum)
	ld	r16,Z+			;Get sum of entire buffer
	ld	r17,Z+
	ld	r18,Z+
	ldi	ZL,LOW(AvgBuffer)
	ldi	ZH,HIGH(AvgBuffer)
	mov	r0,AvgBufferPtr
	lsl	r0
	add	ZL,r0
	clr	r0
	adc	ZH,r0
	ld	YL,Z+			;Get old buffervalue
	ld	YH,Z+
	sub	r16,YL			;Subtract old buffervalue
	sbc	r17,YH
	sbc	r18,r0			;r0=0 !!
	add	r16,XL			;Add new buffervalue
	adc	r17,XH
	adc	r18,r0			;r0=0 !!
	st	-Z,XH			;Save new buffervalue
	st	-Z,XL
	ldi	ZL,LOW(AvgSum)
	ldi	ZH,HIGH(AvgSum)
	st	Z+,r16			;Save sum of entire buffer
	st	Z+,r17
	st	Z+,r18
	ldi	XL,AVG_SIZEEXP		;Prepare to divide
UAB_DivideLoop:
	lsr	r18			;Divide sum to find average pulsewidth
	ror	r17
	ror	r16
	dec	XL
	brne	UAB_DivideLoop
	ldi	ZL,LOW(InAvgPulse)
	ldi	ZH,HIGH(InAvgPulse)
	st	Z+,r16			;Save average pulsewidth
	st	Z+,r17
	inc	AvgBufferPtr
	ldi	r16,AVG_ELEMENTS
	cp	AvgBufferPtr,r16
	brlo	UAB_NoWrap
	clr	AvgBufferPtr
UAB_NoWrap:
	pop	r18
	pop	r17
	pop	r16
	pop	r0
	ret



;**** Main program initialization
RESET:
	ldi	ZL,SRAM_START		;Clear RAM
	clr	ZH
	clr	r16
ClearLoop:
	st	Z+,r16
	cpi	ZL,(RAMEND+1)
	brne	ClearLoop

	ldi	r16,RAMEND
	out	SPL,r16			;Initialize stackpointer

	clr	State			;State init
	ldi	MsCount10,10
	clr	Keyboard
	clr	Flags
	clr	AvgBufferPtr
	ldi	r16,INWATCH_TIME
	mov	InputWatchdog,r16
	ldi	r16,LOW(INIT_PULSE_WIDTH)
	mov	OutPulseL,r16
	ldi	r16,HIGH(INIT_PULSE_WIDTH)
	mov	OutPulseH,r16
	ldi	r16,LOW(INIT_SWEEP_MIN)
	sts	SweepMinPulse,r16
	ldi	r16,HIGH(INIT_SWEEP_MIN)
	sts	SweepMinPulse+1,r16
	ldi	r16,LOW(INIT_SWEEP_MAX)
	sts	SweepMaxPulse,r16
	ldi	r16,HIGH(INIT_SWEEP_MAX)
	sts	SweepMaxPulse+1,r16
	ldi	r16,LOW(INIT_SWEEP_STEP)
	sts	SweepStepPulse,r16
	ldi	r16,HIGH(INIT_SWEEP_STEP)
	sts	SweepStepPulse+1,r16
	
	clr	CursorPos
	rcall	UpdateChangeBy
	ldi	Menu,MENU_OUT

	clr	r16
	out	PORTB,r16
	out	DDRD,r16		;Port D input
	ser	r16
	out	DDRB,r16		;Port B output
	out	PORTD,r16		;Port D pullup on all inputs

	ldi	r16,(1<<CS02)
	out	TCCR0,r16		;Count at CK/256 rate
	ldi	r16,TC0ReloadValue
	out	TCNT0,r16
	ldi	r16,(1<<COM1A1)
	out	TCCR1A,r16		;OC1 output clear on compare match
	ldi	r16,(1<<ICNC1)+(1<<ICES1)+(1<<CS10)
	out	TCCR1B,r16		;Input capture noise canceller, capture at rising edge, count at CK rate
	ldi	r16,(1<<TOIE1)+(1<<ICIE1)+(1<<TOIE0)
	out	TIMSK,r16		;Interrupt at TC0 and TC1 overflow and input capture
	sei				;Enable interrupts

	rcall	LCD_Init		;Interrupts must be enabled before LCD access (T0 used for delay)

	ldi	r16,0x80		;Start of first character
	rcall	LCD_SendCommand		;Display position
	ldi	ZL,InitText1a
	rcall	DisplayText8_r1617_Z
	ldi	r16,0xC0		;Start of 9'th character
	rcall	LCD_SendCommand		;Display position
	ldi	ZL,InitText1b
	rcall	DisplayText8_r1617_Z
	ldi	XL,LOW(2000)
	ldi	XH,HIGH(2000)
	rcall	Delay_ms		;Delay 2 s

	ldi	r16,0x80		;Start of first character
	rcall	LCD_SendCommand		;Display position
	ldi	ZL,InitText2a
	rcall	DisplayText8_r1617_Z
	ldi	r16,0xC0		;Start of 9'th character
	rcall	LCD_SendCommand		;Display position
	ldi	ZL,InitText2b
	rcall	DisplayText8_r1617_Z
	ldi	XL,LOW(2000)
	ldi	XH,HIGH(2000)
	rcall	Delay_ms		;Delay 2 s
	rcall	UpdateDisplay

	rjmp	MainLoop


;**** Program start
MainLoop:
	sbrs	Flags,Flags_NewInput
	rjmp	NoNewInput
	cbr	Flags,(1<<Flags_NewInput)
	rcall	UpdateAverageBuffer
NoNewInput:

	sbrs	Flags,Flags_UpdateDisplay
	rjmp	NoDispUpdate
	cbr	Flags,(1<<Flags_UpdateDisplay)
	rcall	UpdateDisplay
NoDispUpdate:

	;Input from user here
	mov	r16,Keyboard
	sbrs	r16,KEY_UP
	rjmp	NoKeyUp
	ldi	ZL,LOW(KeyUpTable)
	ldi	ZH,HIGH(KeyUpTable)
	rcall	ExecuteKeyCommand_Z
NoKeyUp:
	sbrs	r16,KEY_DOWN
	rjmp	NoKeyDown
	ldi	ZL,LOW(KeyDownTable)
	ldi	ZH,HIGH(KeyDownTable)
	rcall	ExecuteKeyCommand_Z
NoKeyDown:
	sbrs	r16,KEY_LEFT
	rjmp	NoKeyLeft
	ldi	ZL,LOW(KeyLeftTable)
	ldi	ZH,HIGH(KeyLeftTable)
	rcall	ExecuteKeyCommand_Z
NoKeyLeft:
	sbrs	r16,KEY_RIGHT
	rjmp	NoKeyRight
	ldi	ZL,LOW(KeyRightTable)
	ldi	ZH,HIGH(KeyRightTable)
	rcall	ExecuteKeyCommand_Z
NoKeyRight:
	eor	Keyboard,r16		;Clear executed keypress(es)
	tst	r16
	breq	NoKeysPressed
	rcall	UpdateDisplay
NoKeysPressed:

	rjmp	MainLoop		;repeat loop endlessly 
